using System.Windows;
using System.Windows.Media;
using System.Windows.Media.Media3D;

namespace TestProject
{
    /// <summary>
    ///  ־ ̽
    /// </summary>
    public abstract class ModelVisualBase : ModelVisual3D
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        ///  Ű
        /// </summary>
        private static readonly DependencyPropertyKey GeometryKey = DependencyProperty.RegisterReadOnly
        (
            "Geometry",
            typeof(MeshGeometry3D),
            typeof(ModelVisualBase),
            new PropertyMetadata(new MeshGeometry3D())
        );

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Public

        #region Field

        /// <summary>
        /// Ī Ӽ
        /// </summary>
        public static readonly DependencyProperty NameProperty = DependencyProperty.Register
        (
            "Name",
            typeof(string),
            typeof(ModelVisualBase)
        );

        /// <summary>
        ///  Ӽ
        /// </summary>
        public static readonly DependencyProperty MaterialProperty = GeometryModel3D.MaterialProperty.AddOwner
        (
            typeof(ModelVisualBase),
            new PropertyMetadata(MaterialPropertyChangedCallback)
        );

        /// <summary>
        /// ĸ  Ӽ
        /// </summary>
        public static readonly DependencyProperty BackMaterialProperty = GeometryModel3D.BackMaterialProperty.AddOwner
        (
            typeof(ModelVisualBase),
            new PropertyMetadata(MaterialPropertyChangedCallback)
        );

        /// <summary>
        ///  Ӽ
        /// </summary>
        public static readonly DependencyProperty GeometryProperty = GeometryKey.DependencyProperty;

        /// <summary>
        /// ˰ ȯ ÷ Ӽ
        /// </summary>
        public static readonly DependencyProperty AlgorithmicTransformsProperty = DependencyProperty.Register
        (
            "AlgorithmicTransform",
            typeof(AlgorithmicTransformCollection),
            typeof(ModelVisualBase),
            new PropertyMetadata(null, PropertyChangedCallback)
        );

        /// <summary>
        /// ̾   Ӽ
        /// </summary>
        public static readonly DependencyProperty IsWireFrameProperty = DependencyProperty.Register
        (
            "IsWireFrame",
            typeof(bool),
            typeof(ModelVisualBase),
            new PropertyMetadata(false, PropertyChangedCallback)
        );

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        ///   ÷
        /// </summary>
        private Point3DCollection previousVertexCollection = new Point3DCollection();

        /// <summary>
        ///   ÷
        /// </summary>
        private Vector3DCollection previousNormalCollection = new Vector3DCollection();

        /// <summary>
        ///   Ʈ ÷
        /// </summary>
        private Point3DCollection previousNormalPointCollection = new Point3DCollection();

        /// <summary>
        /// ̾ 
        /// </summary>
        private WireFrame wireframe;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region Ī - Name

        /// <summary>
        /// Ī
        /// </summary>
        public string Name
        {
            set
            {
                SetValue(NameProperty, value);
            }
            get
            {
                return (string)GetValue(NameProperty);
            }
        }

        #endregion
        #region  - Material

        /// <summary>
        /// 
        /// </summary>
        public Material Material
        {
            set
            {
                SetValue(MaterialProperty, value);
            }
            get
            {
                return (Material)GetValue(MaterialProperty);
            }
        }

        #endregion
        #region ĸ  - BackMaterial

        /// <summary>
        /// ĸ 
        /// </summary>
        public Material BackMaterial
        {
            set
            {
                SetValue(BackMaterialProperty, value);
            }
            get
            {
                return (Material)GetValue(BackMaterialProperty);
            }
        }

        #endregion
        #region  - Geometry

        /// <summary>
        /// 
        /// </summary>
        public MeshGeometry3D Geometry
        {
            protected set
            {
                SetValue(GeometryKey, value);
            }
            get
            {
                return (MeshGeometry3D)GetValue(GeometryProperty);
            }
        }

        #endregion
        #region ˰ ȯ Ӽ ÷ - AlgorithmicTransforms

        /// <summary>
        /// ˰ ȯ Ӽ ÷
        /// </summary>
        public AlgorithmicTransformCollection AlgorithmicTransforms
        {
            set
            {
                SetValue(AlgorithmicTransformsProperty, value);
            }
            get
            {
                return (AlgorithmicTransformCollection)GetValue(AlgorithmicTransformsProperty);
            }
        }

        #endregion
        #region ̾   - IsWireFrame

        /// <summary>
        /// ̾  
        /// </summary>
        public bool IsWireFrame
        {
            set
            {
                SetValue(IsWireFrameProperty, value);
            }
            get
            {
                return (bool)GetValue(IsWireFrameProperty);
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region  - ModelVisualBase()

        /// <summary>
        /// 
        /// </summary>
        public ModelVisualBase()
        {
            Geometry = Geometry.Clone();

            GeometryModel3D model = new GeometryModel3D(Geometry, null);

            Content = model;

            AlgorithmicTransforms = new AlgorithmicTransformCollection();
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region  Ӽ  ݹ óϱ - MaterialPropertyChangedCallback(d, e)

        /// <summary>
        ///  Ӽ  ݹ óϱ
        /// </summary>
        /// <param name="d"> ü</param>
        /// <param name="e">̺Ʈ </param>
        private static void MaterialPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            (d as ModelVisualBase).MaterialPropertyChanged(e);
        }

        #endregion
        #region Ӽ  ݹ óϱ - PropertyChangedCallback(d, e)

        /// <summary>
        /// Ӽ  ݹ óϱ
        /// </summary>
        /// <param name="d"> ü</param>
        /// <param name="e">̺Ʈ </param>
        protected static void PropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            (d as ModelVisualBase).PropertyChanged(e);
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Protected

        #region ﰢ ϱ - Triangulate(e, vertexCollection, normalCollection, indexCollection, textureCollection)

        /// <summary>
        /// ﰢ ϱ
        /// </summary>
        /// <param name="e">̺Ʈ </param>
        /// <param name="vertexCollection"> ÷</param>
        /// <param name="normalCollection"> ÷</param>
        /// <param name="indexCollection">ε ÷</param>
        /// <param name="textureCollection">ؽó ÷</param>
        protected abstract void Triangulate(DependencyPropertyChangedEventArgs e, Point3DCollection vertexCollection, Vector3DCollection normalCollection, Int32Collection indexCollection, PointCollection textureCollection);

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Private

        #region  Ӽ  óϱ - MaterialPropertyChanged(e)

        /// <summary>
        ///  Ӽ  óϱ
        /// </summary>
        /// <param name="e">̺Ʈ </param>
        private void MaterialPropertyChanged(DependencyPropertyChangedEventArgs e)
        {
            GeometryModel3D model = Content as GeometryModel3D;

            if(e.Property == MaterialProperty)
            {
                model.Material = e.NewValue as Material;
            }
            else if(e.Property == BackMaterialProperty)
            {
                model.BackMaterial = e.NewValue as Material;
            }
        }

        #endregion
        #region Ӽ  óϱ - PropertyChanged(e)

        /// <summary>
        /// Ӽ  óϱ
        /// </summary>
        /// <param name="e">̺Ʈ </param>
        protected virtual void PropertyChanged(DependencyPropertyChangedEventArgs e)
        {
            Point3DCollection  vertexCollection;
            Vector3DCollection normalCollection;
            Int32Collection    indexCollection;
            PointCollection    textureCollection;

            if(!IsWireFrame || e.Property == IsWireFrameProperty && !(bool)e.OldValue)
            {
                MeshGeometry3D mesh = Geometry;

                vertexCollection  = mesh.Positions;
                normalCollection  = mesh.Normals;
                indexCollection   = mesh.TriangleIndices;
                textureCollection = mesh.TextureCoordinates;

                mesh.Positions          = null;
                mesh.Normals            = null;
                mesh.TriangleIndices    = null;
                mesh.TextureCoordinates = null;
            }
            else
            {
                vertexCollection  = this.wireframe.Positions;
                normalCollection  = this.wireframe.Normals;
                indexCollection   = this.wireframe.TriangleIndices;
                textureCollection = this.wireframe.TextureCoordinates;

                this.wireframe.Positions          = null;
                this.wireframe.Normals            = null;
                this.wireframe.TriangleIndices    = null;
                this.wireframe.TextureCoordinates = null;
            }

            if(e.Property != AlgorithmicTransformsProperty && e.Property != IsWireFrameProperty)
            {
                Triangulate(e, vertexCollection, normalCollection, indexCollection, textureCollection);

                this.previousVertexCollection.Clear();

                this.previousNormalCollection.Clear();

                foreach(Point3D vertex in vertexCollection)
                {
                    this.previousVertexCollection.Add(vertex);
                }

                foreach(Vector3D normal in normalCollection)
                {
                    this.previousNormalCollection.Add(normal);
                }
            }

            if(e.Property == AlgorithmicTransformsProperty)
            {
                vertexCollection.Clear();
                normalCollection.Clear();
                this.previousNormalPointCollection.Clear();

                foreach(Point3D vertex in this.previousVertexCollection)
                {
                    vertexCollection.Add(vertex);
                }

                foreach(Vector3D normal in this.previousNormalCollection)
                {
                    this.previousNormalPointCollection.Add((Point3D)normal);
                }
            }

            if(e.Property != IsWireFrameProperty)
            {
                foreach(AlgorithmicTransform transform in AlgorithmicTransforms)
                {
                    transform.Transform(vertexCollection);

                    transform.Transform(this.previousNormalPointCollection);
                }

                foreach(Point3D point in this.previousNormalPointCollection)
                {
                    normalCollection.Add((Vector3D)point);
                }
            }

            if(IsWireFrame)
            {
                if(this.wireframe == null)
                {
                    this.wireframe = new WireFrame();

                    Children.Add(this.wireframe);
                }

                this.wireframe.TextureCoordinates = textureCollection;
                this.wireframe.TriangleIndices    = indexCollection;
                this.wireframe.Normals            = normalCollection;
                this.wireframe.Positions          = vertexCollection;
            }
            else
            {
                MeshGeometry3D mesh = Geometry;

                mesh.TextureCoordinates = textureCollection;
                mesh.TriangleIndices    = indexCollection;
                mesh.Normals            = normalCollection;
                mesh.Positions          = vertexCollection;
            }
        }

        #endregion
    }
}