using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Media;
using System.Windows.Media.Media3D;

namespace TestProject
{
    /// <summary>
    /// ̾ ̽
    /// </summary>
    public abstract class WireBase : ModelVisual3D
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// ̾ ̽  Ʈ
        /// </summary>
        private static List<WireBaseInfo> _wireBaseInfoList = new List<WireBaseInfo>();

        /// <summary>
        ///  ̾ ̽  Ʈ
        /// </summary>
        private static List<WireBaseInfo> _removeWireBaseInfoList = new List<WireBaseInfo>();

        /// <summary>
        ///  ÷ Ű
        /// </summary>
        private static readonly DependencyPropertyKey LineCollectionKey = DependencyProperty.RegisterReadOnly
        (
            "LineCollection",
            typeof(Point3DCollection),
            typeof(WireBase),
            new PropertyMetadata(null, RecalculcationPropertyChangedCallback)
        );

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Public

        #region Field

        /// <summary>
        ///  ÷ Ӽ
        /// </summary>
        public static readonly DependencyProperty LineCollectionProperty = LineCollectionKey.DependencyProperty;

        /// <summary>
        ///  Ӽ
        /// </summary>
        public static readonly DependencyProperty ColorProperty = DependencyProperty.Register
        (
            "Color",
            typeof(Color),
            typeof(WireBase),
            new PropertyMetadata(Colors.Black, ColorPropertyChangedCallback)
        );

        /// <summary>
        /// β Ӽ
        /// </summary>
        public static readonly DependencyProperty ThicknessProperty = DependencyProperty.Register
        (
            "Thickness",
            typeof(double),
            typeof(WireBase),
            new PropertyMetadata(1.0, RecalculcationPropertyChangedCallback)
        );

        /// <summary>
        ///  Ӽ
        /// </summary>
        public static readonly DependencyProperty RoundingProperty = DependencyProperty.Register
        (
            "Rounding",
            typeof(int),
            typeof(WireBase),
            new PropertyMetadata(0, RecalculcationPropertyChangedCallback)
        );

        /// <summary>
        /// ȭǥ  Ӽ
        /// </summary>
        public static readonly DependencyProperty ArrowAngleProperty = DependencyProperty.Register
        (
            "ArrowAngle",
            typeof(double),
            typeof(WireBase),
            new PropertyMetadata(45.0, RecalculcationPropertyChangedCallback)
        );

        /// <summary>
        /// ȭǥ  Ӽ
        /// </summary>
        public static readonly DependencyProperty ArrowLengthProperty = DependencyProperty.Register
        (
            "ArrowLength",
            typeof(double),
            typeof(WireBase),
            new PropertyMetadata(12.0, RecalculcationPropertyChangedCallback)
        );

        /// <summary>
        /// ȭǥ Ÿ Ӽ
        /// </summary>
        public static readonly DependencyProperty ArrowTypeProperty = DependencyProperty.Register
        (
            "ArrowType",
            typeof(ArrowType),
            typeof(WireBase),
            new PropertyMetadata(ArrowType.None, RecalculcationPropertyChangedCallback)
        );

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        ///  ʿ 
        /// </summary>
        private bool needRecalculation = true;

        /// <summary>
        /// ־ȭ Ʈ
        /// </summary>
        private Matrix3D visualToScreenMatrix = Matrix3D.Identity;

        /// <summary>
        /// ȭ־ Ʈ
        /// </summary>
        private Matrix3D screenToVisualMatrix;

        /// <summary>
        /// ȸ ȯ
        /// </summary>
        private RotateTransform rotateTransform = new RotateTransform();

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region  ÷ - LineCollection

        /// <summary>
        ///  ÷
        /// </summary>
        public Point3DCollection LineCollection
        {
            private set
            {
                SetValue(LineCollectionKey, value);
            }
            get
            {
                return (Point3DCollection)GetValue(LineCollectionProperty);
            }
        }

        #endregion
        #region  - Color

        /// <summary>
        /// 
        /// </summary>
        public Color Color
        {
            set
            {
                SetValue(ColorProperty, value);
            }
            get
            {
                return (Color)GetValue(ColorProperty);
            }
        }

        #endregion
        #region β - Thickness

        /// <summary>
        /// β
        /// </summary>
        public double Thickness
        {
            set
            {
                SetValue(ThicknessProperty, value);
            }
            get
            {
                return (double)GetValue(ThicknessProperty);
            }
        }

        #endregion
        #region  - Rounding

        /// <summary>
        /// 
        /// </summary>
        public int Rounding
        {
            set
            {
                SetValue(RoundingProperty, value);
            }
            get
            {
                return (int)GetValue(RoundingProperty);
            }
        }

        #endregion
        #region ȭǥ  - ArrowAngle

        /// <summary>
        /// ȭǥ 
        /// </summary>
        public double ArrowAngle
        {
            set
            {
                SetValue(ArrowAngleProperty, value);
            }
            get
            {
                return (double)GetValue(ArrowAngleProperty);
            }
        }

        #endregion
        #region ȭǥ  - ArrowLength

        /// <summary>
        /// ȭǥ 
        /// </summary>
        public double ArrowLength
        {
            set
            {
                SetValue(ArrowLengthProperty, value);
            }
            get
            {
                return (double)GetValue(ArrowLengthProperty);
            }
        }

        #endregion
        #region ȭǥ Ÿ - ArrowType

        /// <summary>
        /// ȭǥ Ÿ
        /// </summary>
        public ArrowType ArrowType
        {
            set
            {
                SetValue(ArrowTypeProperty, value);
            }
            get
            {
                return (ArrowType)GetValue(ArrowTypeProperty);
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Static

        #region  - WireBase()

        /// <summary>
        /// 
        /// </summary>
        static WireBase()
        {
            CompositionTarget.Rendering += CompositionTarget_Rendering;
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Public

        #region  - WireBase()

        /// <summary>
        /// 
        /// </summary>
        public WireBase()
        {
            LineCollection = new Point3DCollection();

            MeshGeometry3D mesh = new MeshGeometry3D();

            MaterialGroup materialGroup = new MaterialGroup();

            materialGroup.Children.Add(new DiffuseMaterial(Brushes.Black));
            materialGroup.Children.Add(new EmissiveMaterial(new SolidColorBrush(Color)));

            GeometryModel3D model = new GeometryModel3D(mesh, materialGroup);

            model.BackMaterial = new DiffuseMaterial(Brushes.Red);

            Content = model;

            _wireBaseInfoList.Add(new WireBaseInfo(this));
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region  Ӽ  ݹ óϱ - RecalculcationPropertyChangedCallback(d, e)

        /// <summary>
        ///  Ӽ  ݹ óϱ
        /// </summary>
        /// <param name="d"> ü</param>
        /// <param name="e">̺Ʈ </param>
        private static void RecalculcationPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            WireBase wireBase = d as WireBase;

            wireBase.needRecalculation = true;
        }

        #endregion
        #region  Ӽ  ݹ óϱ - ColorPropertyChangedCallback(d, e)

        /// <summary>
        ///  Ӽ  ݹ óϱ
        /// </summary>
        /// <param name="d"> ü</param>
        /// <param name="e">̺Ʈ </param>
        private static void ColorPropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            WireBase wirebase = d as WireBase;

            GeometryModel3D model = wirebase.Content as GeometryModel3D;

            MaterialGroup group = model.Material as MaterialGroup;

            EmissiveMaterial material = group.Children[1] as EmissiveMaterial;

            material.Brush = new SolidColorBrush((Color)e.NewValue);
        }

        #endregion
        #region Ӽ  ݹ óϱ - PropertyChangedCallback(d, e)

        /// <summary>
        /// Ӽ  ݹ óϱ
        /// </summary>
        /// <param name="d"> ü</param>
        /// <param name="e">̺Ʈ </param>
        protected static void PropertyChangedCallback(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            (d as WireBase).PropertyChanged(e);
        }

        #endregion

        #region ռ Ÿ  óϱ - CompositionTarget_Rendering(sender, e)

        /// <summary>
        /// ռ Ÿ  óϱ
        /// </summary>
        /// <param name="sender">̺Ʈ ߻</param>
        /// <param name="e">̺Ʈ </param>
        private static void CompositionTarget_Rendering(object sender, EventArgs e)
        {
            foreach(WireBaseInfo wireBaseInfo in _wireBaseInfoList)
            {
                WireBase wireBase = wireBaseInfo.WireBase;

                DependencyObject d = wireBase;

                while(d != null && !(d is Window))
                {
                    d = VisualTreeHelper.GetParent(d);
                }

                if(wireBaseInfo.Window == null)
                {
                    if(d != null)
                    {
                        wireBaseInfo.Window = d as Window;
                    }
                    else
                    {
                        continue;
                    }
                }
                else
                {
                    if(d == null)
                    {
                        _removeWireBaseInfoList.Add(wireBaseInfo);

                        continue;
                    }
                }

                wireBase.OnRendering();
            }

            if(_removeWireBaseInfoList.Count > 0)
            {
                foreach(WireBaseInfo wireBaseInfo in _removeWireBaseInfoList)
                {
                    _wireBaseInfoList.Remove(wireBaseInfo);
                }

                _removeWireBaseInfoList.Clear();
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Protected

        #region Ӽ  óϱ - PropertyChanged(e)

        /// <summary>
        /// Ӽ  óϱ
        /// </summary>
        /// <param name="e">̺Ʈ </param>
        protected void PropertyChanged(DependencyPropertyChangedEventArgs e)
        {
            Point3DCollection collection = LineCollection;

            LineCollection = null;

            Generate(e, collection);

            LineCollection = collection;
        }

        #endregion
        #region ϱ - Generate(e, lineCollection)

        /// <summary>
        /// ϱ
        /// </summary>
        /// <param name="e">̺Ʈ </param>
        /// <param name="lineCollection"> ÷</param>
        protected abstract void Generate(DependencyPropertyChangedEventArgs e, Point3DCollection lineCollection);

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Private

        #region  - Widen(inPoint3D, deltaVector)

        /// <summary>
        /// 
        /// </summary>
        /// <param name="inPoint3D">Է Ʈ 3D</param>
        /// <param name="deltaVector">Ÿ </param>
        /// <returns>Ʈ 3D</returns>
        private Point3D Widen(Point3D inPoint3D, Vector deltaVector)
        {
            Point4D inPoint4D  = (Point4D)inPoint3D;
            Point4D outPoint4D = inPoint4D * visualToScreenMatrix;

            outPoint4D.X += deltaVector.X * outPoint4D.W;
            outPoint4D.Y += deltaVector.Y * outPoint4D.W;

            outPoint4D *= screenToVisualMatrix;

            Point3D outPoint3D = new Point3D
            (
                outPoint4D.X / outPoint4D.W,
                outPoint4D.Y / outPoint4D.W,
                outPoint4D.Z / outPoint4D.W
            );

            return outPoint3D;
        }

        #endregion
        #region Ʈ ϱ - CalculatePoint(inPoint3D, deltaVector, angle)

        /// <summary>
        /// Ʈ ϱ
        /// </summary>
        /// <param name="inPoint3D">Է Ʈ 3D</param>
        /// <param name="deltaVector">Ÿ </param>
        /// <param name="angle"></param>
        /// <returns>Ʈ 3D</returns>
        private Point3D CalculatePoint(Point3D inPoint3D, Vector deltaVector, double angle)
        {
            Point4D inPoint4D  = (Point4D)inPoint3D;
            Point4D outPoint4D = inPoint4D * visualToScreenMatrix;

            this.rotateTransform.Angle = angle;

            deltaVector = (Vector)this.rotateTransform.Transform((Point)deltaVector);

            outPoint4D.X += deltaVector.X * outPoint4D.W;
            outPoint4D.Y += deltaVector.Y * outPoint4D.W;

            outPoint4D *= screenToVisualMatrix;

            Point3D outPoint3D = new Point3D
            (
                outPoint4D.X / outPoint4D.W,
                outPoint4D.Y / outPoint4D.W,
                outPoint4D.Z / outPoint4D.W
            );

            return outPoint3D;
        }

        #endregion
        #region  ߰ϱ - AddRounding(inPoint3D, deltaVector, pointCollection, indexCollection, indexCount)

        /// <summary>
        ///  ߰ϱ
        /// </summary>
        /// <param name="inPoint3D">Է Ʈ 3D</param>
        /// <param name="deltaVector">Ÿ </param>
        /// <param name="pointCollection">Ʈ ÷</param>
        /// <param name="indexCollection">ε ÷</param>
        /// <param name="indexCount">ε īƮ</param>
        private void AddRounding(Point3D inPoint3D, Vector deltaVector, Point3DCollection pointCollection, Int32Collection indexCollection, ref int indexCount)
        {
            pointCollection.Add(CalculatePoint(inPoint3D, new Vector(0, 0), 0));

            for(int i = 0; i <= Rounding; i++)
            {
                pointCollection.Add(CalculatePoint(inPoint3D, deltaVector, 180 * i / Rounding));
            }

            for(int i = 0; i < Rounding; i++)
            {
                indexCollection.Add(indexCount        );
                indexCollection.Add(indexCount + i + 2);
                indexCollection.Add(indexCount + i + 1);
            }

            indexCount += Rounding + 2;
        }

        #endregion
        #region  óϱ - ProcessLine(point1, point2, pointCollection, indexCollection, indexBase)

        /// <summary>
        ///  óϱ
        /// </summary>
        /// <param name="point1">Ʈ 1</param>
        /// <param name="point2">Ʈ 2</param>
        /// <param name="pointCollection">Ʈ ÷</param>
        /// <param name="indexCollection">ε ÷</param>
        /// <param name="indexBase">ε ̽</param>
        private void ProcessLine(Point3D point1, Point3D point2, Point3DCollection pointCollection, Int32Collection indexCollection, ref int indexBase)
        {
            Point3D screenPoint1 = point1 * visualToScreenMatrix;
            Point3D screenPoint2 = point2 * visualToScreenMatrix;

            Vector3D lineVector = screenPoint2 - screenPoint1;

            lineVector.Z = 0;

            lineVector.Normalize();

            Vector deltaVector = (Thickness / 2) * new Vector(-lineVector.Y, lineVector.X);

            pointCollection.Add(Widen(point1,  deltaVector));
            pointCollection.Add(Widen(point1, -deltaVector));
            pointCollection.Add(Widen(point2,  deltaVector));
            pointCollection.Add(Widen(point2, -deltaVector));

            indexCollection.Add(indexBase    );
            indexCollection.Add(indexBase + 2);
            indexCollection.Add(indexBase + 1);
            indexCollection.Add(indexBase + 1);
            indexCollection.Add(indexBase + 2);
            indexCollection.Add(indexBase + 3);

            indexBase += 4;

            if(Rounding > 0)
            {
                AddRounding(point1,  deltaVector, pointCollection, indexCollection, ref indexBase);
                AddRounding(point2, -deltaVector, pointCollection, indexCollection, ref indexBase);
            }
        }

        #endregion
        #region ȭǥ óϱ - ProcessArrow(point1, point2, pointCollection, indexCollection, indexBase)

        /// <summary>
        /// ȭǥ óϱ
        /// </summary>
        /// <param name="point1">Ʈ 1</param>
        /// <param name="point2">Ʈ 2</param>
        /// <param name="pointCollection">Ʈ ÷</param>
        /// <param name="indexCollection">ε ÷</param>
        /// <param name="indexBase">ε ̽</param>
        private void ProcessArrow(Point3D point1, Point3D point2, Point3DCollection pointCollection, Int32Collection indexCollection, ref int indexBase)
        {
            Point3D screenPoint1 = point1 * visualToScreenMatrix;
            Point3D screenPoint2 = point2 * visualToScreenMatrix;

            Vector arrowVector = new Vector(screenPoint1.X - screenPoint2.X, screenPoint1.Y - screenPoint2.Y);

            arrowVector.Normalize();

            arrowVector *= ArrowLength;

            Matrix matrix = new Matrix();

            matrix.Rotate(ArrowAngle / 2);

            Point3D arrowPoint1 = Widen(point2, arrowVector * matrix);

            matrix.Rotate(-ArrowAngle);

            Point3D arrowPoint2 = Widen(point2, arrowVector * matrix);

            ProcessLine(point2, arrowPoint1, pointCollection, indexCollection, ref indexBase);
            ProcessLine(point2, arrowPoint2, pointCollection, indexCollection, ref indexBase);
        }

        #endregion
        #region ٽ ϱ - Recalculate()

        /// <summary>
        /// ٽ ϱ
        /// </summary>
        private void Recalculate()
        {
            GeometryModel3D model = Content as GeometryModel3D;

            MeshGeometry3D mesh = model.Geometry as MeshGeometry3D;

            Point3DCollection pointCollection = mesh.Positions;

            mesh.Positions = null;

            pointCollection.Clear();

            Int32Collection indexCollection = mesh.TriangleIndices;

            mesh.TriangleIndices = null;

            indexCollection.Clear();

            int indexBase = 0;

            Point3DCollection lineCollection = LineCollection;

            for(int line = 0; line < lineCollection.Count - 1; line += 2)
            {
                Point3D point1 = lineCollection[line + 0];
                Point3D point2 = lineCollection[line + 1];

                ProcessLine(point1, point2, pointCollection, indexCollection, ref indexBase);

                if(line == 0 && (ArrowType & ArrowType.Start) == ArrowType.Start)
                {
                    ProcessArrow(point2, point1, pointCollection, indexCollection, ref indexBase);
                }

                if(line > lineCollection.Count - 4 && (ArrowType & ArrowType.End) == ArrowType.End)
                {
                    ProcessArrow(point1, point2, pointCollection, indexCollection, ref indexBase);
                }
            }

            mesh.TriangleIndices = indexCollection;

            mesh.Positions = pointCollection;
        }

        #endregion
        #region  óϱ - OnRendering()

        /// <summary>
        ///  óϱ
        /// </summary>
        private void OnRendering()
        {
            if(LineCollection.Count == 0)
            {
                return;
            }

            Matrix3D matrix = VisualInfo.GetTotalTransform(this);

            if(matrix == VisualInfo.ZeroMatrix)
            {
                return;
            }

            if(matrix.IsIdentity)
            {
                return;
            }

            if(matrix != visualToScreenMatrix)
            {
                this.visualToScreenMatrix = matrix;
                this.screenToVisualMatrix = matrix;

                if(this.screenToVisualMatrix.HasInverse)
                {
                    this.screenToVisualMatrix.Invert();

                    this.needRecalculation = true;
                }
                else
                {
                    throw new ApplicationException("Here's where the problem is");
                }
            }

            if(this.needRecalculation)
            {
                Recalculate();

                this.needRecalculation = false;
            }
        }

        #endregion
    }
}